/*
 * (C) Copyright 2003
 * Kyle Harris, kharris@nexus-tech.net
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <command.h>
#include <mmc.h>
#include <malloc.h>

extern int mmc_switch(struct mmc *mmc, u8 set, u8 index, u8 value);
extern int mmc_send_status(struct mmc *mmc, int timeout,int no_crc);
extern int export_mmc_show_ext_csd(unsigned int idx);
extern int export_mmc_switch_part(unsigned int part_num);

static int curr_device = -1;

static struct mmc *init_mmc_device(int dev, bool force_init)
{
	struct mmc *mmc;
	mmc = find_mmc_device(dev);
	if (!mmc) {
		printf("no mmc device at slot %x\n", dev);
		return NULL;
	}

	if (force_init){
		mmc->has_init = 0;
		mmc->clock = 400000;
		mmc->bus_width = 0;
	}
	if (mmc_init(mmc))
		return NULL;
	return mmc;
}

enum mmc_state {
	MMC_INVALID,
	MMC_READ,
	MMC_WRITE,
	MMC_ERASE,
#ifdef CONFIG_RTK_MMC
	MMC_READ_BYTE,
	MMC_WRITE_BYTE,
	MMC_ERASE_SECU,
	MMC_ERASE_TRIM,
#endif
};

#ifdef CONFIG_RTK_MMC
extern int mmc_send_cmd(struct mmc *mmc, struct mmc_cmd *cmd, struct mmc_data *data);
extern u64 rtk_eMMC_read_byte(u64  byte_addr_str,unsigned int data_size,unsigned int * buffer );
extern u64 rtk_eMMC_write_byte(u64  byte_addr_str,unsigned int data_size,unsigned int * buffer );
#endif

static void print_mmcinfo(struct mmc *mmc)
{
	printf("Device: %s\n", mmc->name);
	printf("Manufacturer ID: %x\n", mmc->cid[0] >> 24);
	printf("OEM: %x\n", (mmc->cid[0] >> 8) & 0xffff);
	printf("Name: %c%c%c%c%c \n", mmc->cid[0] & 0xff,
			(mmc->cid[1] >> 24), (mmc->cid[1] >> 16) & 0xff,
			(mmc->cid[1] >> 8) & 0xff, mmc->cid[1] & 0xff);

	printf("Tran Speed: %d\n", mmc->tran_speed);
	printf("Rd Block Len: %d\n", mmc->read_bl_len);

	printf("%s version %d.%d\n", IS_SD(mmc) ? "SD" : "MMC",
			(mmc->version >> 4) & 0xf, mmc->version & 0xf);

	printf("High Capacity: %s\n", mmc->high_capacity ? "Yes" : "No");
	puts("Capacity: ");
	print_size(mmc->capacity, "\n");

	printf("Bus Width: %d-bit\n", mmc->bus_width);
}

int do_mmcinfo (cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	struct mmc *mmc;

	mmc = init_mmc_device(curr_device, false);
	if (!mmc){
		printf("no mmc device at slot %x\n", curr_device);
		return CMD_RET_FAILURE;
	}

	print_mmcinfo(mmc);
	return CMD_RET_SUCCESS;

}

static int do_mmc_read(cmd_tbl_t *cmdtp, int flag,
		       int argc, char * const argv[])
{
	struct mmc *mmc;
	u32 blk;
	u32 cnt, n;
	void *addr;

	if (argc != 4)
		return CMD_RET_USAGE;

	addr = (void *)simple_strtoul(argv[1], NULL, 16);
	blk = simple_strtoul(argv[2], NULL, 16);
	cnt = simple_strtoul(argv[3], NULL, 16);

	mmc = init_mmc_device(curr_device, false);
	if (!mmc){
		printf("no mmc device at slot %x\n", curr_device);
		return CMD_RET_FAILURE;
	}

	printf("\nMMC read: dev # %d, block # %d, count %d ... ",
			curr_device, (u32)blk, cnt);

	n = mmc->block_dev.block_read(curr_device, (u32)blk,cnt, addr);
	/* flush cache after read */
	flush_cache((ulong)addr, cnt * 512); /* FIXME */
	printf("%d blocks read: %s\n", n, (n == cnt) ? "OK" : "ERROR");

	return (n == cnt) ? CMD_RET_SUCCESS : CMD_RET_FAILURE;
}

static int do_mmc_write(cmd_tbl_t *cmdtp, int flag,
			int argc, char * const argv[])
{
	struct mmc *mmc;
	u32 blk;
	u32 cnt, n;
	void *addr;

	if (argc != 4)
		return CMD_RET_USAGE;

	addr = (void *)simple_strtoul(argv[1], NULL, 16);
	blk = simple_strtoul(argv[2], NULL, 16);
	cnt = simple_strtoul(argv[3], NULL, 16);

	mmc = init_mmc_device(curr_device, false);
	if (!mmc){
		printf("no mmc device at slot %x\n", curr_device);
		return CMD_RET_FAILURE;
	}

	printf("\nMMC write: dev # %d, block # %d, count %d ... ",
				curr_device, (u32)blk, cnt);

#if 0	/* should add to check write protect func */
	if (mmc_getwp(mmc) == 1) {
		printf("Error: card is write protected!\n");
		return CMD_RET_FAILURE;
	}
#endif
	n = mmc->block_dev.block_write(curr_device, blk,cnt, addr);
	printf("%d blocks written: %s\n", n, (n == cnt) ? "OK" : "ERROR");

	return (n == cnt) ? CMD_RET_SUCCESS : CMD_RET_FAILURE;
}


static
lbaint_t mmc_erase(struct mmc *mmc, unsigned long start, lbaint_t blkcnt,unsigned int type);
static int do_mmc_erase(cmd_tbl_t *cmdtp, int flag,
			int argc, char * const argv[])
{
	struct mmc *mmc;
	u32 blk, cnt, n=-1;

	if ((argc < 3) || (argc > 4))
		return CMD_RET_USAGE;

	mmc = init_mmc_device(curr_device, false);
	if (!mmc){
		printf("no mmc device at slot %x\n", curr_device);
		return CMD_RET_FAILURE;
	}

	if (argc == 3){
		blk = (u32)simple_strtoul(argv[1], NULL, 16);
		cnt = (u32)simple_strtoul(argv[2], NULL, 16);
		printf("\nMMC erase: dev # %d, block # %d, count %d ... ",
	       			curr_device, blk, cnt);
		n = mmc->block_dev.block_erase(curr_device, blk, cnt);
#ifdef CONFIG_RTK_MMC
	}else{
		blk = (u32)simple_strtoul(argv[2], NULL, 16);
		cnt = (u32)simple_strtoul(argv[3], NULL, 16);
		printf("\nMMC erase: dev # %d, block # %d, count %d ... ",
	       			curr_device, blk, cnt);
		if(strcmp(argv[1], "secure") == 0){
			n = mmc_erase(mmc, (unsigned long)blk, (lbaint_t)cnt, SECURE_ERASE);
		}
		if(strcmp(argv[1], "trim") == 0){
			n = mmc_erase(mmc, (unsigned long)blk, (lbaint_t)cnt, TRIM_OPER);
		}
#endif
	}

	printf("%d blocks erased: %s\n", n, (n == cnt) ? "OK" : "ERROR");

	return (n == cnt) ? CMD_RET_SUCCESS : CMD_RET_FAILURE;
}

static int do_mmc_rescan(cmd_tbl_t *cmdtp, int flag,
			 int argc, char * const argv[])
{
	struct mmc *mmc;

	mmc = init_mmc_device(curr_device, true);
	if (!mmc){
		printf("no mmc device at slot %x\n", curr_device);
		return CMD_RET_FAILURE;
	}

	return CMD_RET_SUCCESS;
}

static int do_mmc_part(cmd_tbl_t *cmdtp, int flag,
			 int argc, char * const argv[])
{
	struct mmc *mmc;
	block_dev_desc_t *mmc_dev;

	mmc = init_mmc_device(curr_device, false);
	if (!mmc)
		return CMD_RET_FAILURE;

	mmc_dev = mmc_get_dev(curr_device);
	if ((mmc_dev != NULL) &&
	    (mmc_dev->type != DEV_TYPE_UNKNOWN))
	{
		print_part(mmc_dev);
		return CMD_RET_SUCCESS;
	}

	puts("get mmc type error!\n");
	return CMD_RET_FAILURE;

}


static int do_mmc_list(cmd_tbl_t *cmdtp, int flag,
			 int argc, char * const argv[])
{
	struct mmc *mmc;

	mmc = init_mmc_device(curr_device, false);
	if (!mmc)
		return CMD_RET_FAILURE;

	print_mmc_devices('\n');
	return CMD_RET_SUCCESS;
}


static int do_mmc_dev(cmd_tbl_t *cmdtp, int flag,
			 int argc, char * const argv[])
{
	int dev, part = -1;
	struct mmc *mmc;

	mmc = init_mmc_device(curr_device, false);
	if (!mmc)
		return CMD_RET_FAILURE;

	if (argc == 1)
		dev = curr_device;
	else if (argc == 2)
		dev = simple_strtoul(argv[2], NULL, 10);
	else if (argc == 3) {
		dev = (int)simple_strtoul(argv[2], NULL, 10);
		part = (int)simple_strtoul(argv[3], NULL, 10);
		if (part > PART_ACCESS_MASK) {
			printf("#part_num shouldn't be larger"
				" than %d\n", PART_ACCESS_MASK);
			return CMD_RET_FAILURE;
		}
	} else
		return CMD_RET_USAGE;

	if (part != -1) {
		int ret;
		if (mmc->part_config == MMCPART_NOAVAILABLE) {
			printf("Card doesn't support part_switch\n");
			return CMD_RET_FAILURE;
		}

		if (part != mmc->part_num) {
			ret = mmc_switch_part(dev, part);
			if (!ret)
				mmc->part_num = part;

			printf("switch to partions #%d, %s\n",
					part, (!ret) ? "OK" : "ERROR");
		}
	}
	curr_device = dev;
	if (mmc->part_config == MMCPART_NOAVAILABLE)
		printf("mmc%d is current device\n", curr_device);
	else
		printf("mmc%d(part %d) is current device\n",
			curr_device, mmc->part_num);

	return CMD_RET_SUCCESS;

}


#ifdef CONFIG_RTK_MMC

#define START_TEST_MMC_BLK 0x10000
#define MMC_RW_TEST_BLK 0x400

#define MMC_RW_PATTERN0 0xf0f0f0f0
#define MMC_RW_PATTERN1 0x0f0f0f0f
#define MMC_RW_PATTERN2 0x5a5a5a5a
#define MMC_RW_PATTERN3 0xa5a5a5a5

/* event control */
#define SHOW_MORE_LOG   (0x01U)
#define SHOW_RESULT     (0x01U<<1)

#define RESULT_DISPLAY  (0x01U)
#define RESULT_RECORD   (0x01U<<1)
#define RESULT_BUF_LEN  (128)


u32*  mmc_test_buf = NULL;
static int do_mmc_life_test(cmd_tbl_t *cmdtp, int flag,
		       int argc, char * const argv[])
{
    int i,j,k;
    int ret;
    u32* w_buf = NULL;
    u32* r_buf = NULL;
    u32 test_loop = 0;
    u32 start_t_blk = 0;
    u32 total_t_len = 0;
    u32 remain_t_len = 0;
    u32 event_ctl = 0;
    u32 n;

    u32 act_blk = 0;
    u32 act_len = 0;
    u32 pattern_ary[4] = {MMC_RW_PATTERN0,MMC_RW_PATTERN1,MMC_RW_PATTERN2,MMC_RW_PATTERN3};
    u32 pattern;
    u32 comp_fail = 0;
    char result_buf[RESULT_BUF_LEN];
    struct mmc *mmc;
    const char *old_result = getenv("em_lf_rst");

    if(strcmp(argv[1], "result") == 0){
        event_ctl = SHOW_RESULT;
        if(old_result){
            printf("\n\nLast MMC life testing result : %s\n\n",old_result);
        }else{
            printf("\n\nHave no result recorded\n\n");
        }
        return 0;
    }else{
        test_loop = simple_strtoul(argv[1], NULL, 10);
        printf("\n\nTest_loop=%d\n",test_loop);
    }

    if(strcmp(argv[2], "show") == 0){
        event_ctl = SHOW_MORE_LOG;
    }

	mmc = init_mmc_device(curr_device, false);
	if (!mmc)
		return CMD_RET_FAILURE;

    if(!mmc_test_buf){
        /* get more buffer for test; first half for write, bottom falf for read  */
		mmc_test_buf = (u32 *)malloc(((MMC_RW_TEST_BLK*2)*512) + 0x20);

        if(!mmc_test_buf){
            printf("get buff fail!\n");
            return 0;
        }
        //printf("mmc_test_buf at 0x%x\n",(u32)mmc_test_buf);
        mmc_test_buf = (u32 *)((((u32)(mmc_test_buf))&(~0x1F)) + 0x20);

    }

    w_buf = mmc_test_buf;
    r_buf = mmc_test_buf+(MMC_RW_TEST_BLK*(512/4));

    //printf("%s(%d)mmc_test_buf at 0x%x\n",__func__,__LINE__,(u32)mmc_test_buf);
    //printf("%s(%d)w_buf at  0x%x; r_buf at 0x%x\n",__func__,__LINE__,(u32)w_buf,(u32)r_buf);
    //printf("%s(%d)mmc->capacity =0x%llx(bytes in long long case)\n",__func__,__LINE__,mmc->capacity );
    //printf("%s(%d)mmc->capacity =0x%llx(block in long long case)\n",__func__,__LINE__,(mmc->capacity >> 9));

    start_t_blk = START_TEST_MMC_BLK;
#if 0
#define FAKE_MMC_CAPACITY (START_TEST_MMC_BLK+0x5000)
    printf("To set fake capacity as 0x%08x for testing !!!\n",FAKE_MMC_CAPACITY);
    total_t_len = FAKE_MMC_CAPACITY - START_TEST_MMC_BLK;
#else
    total_t_len = (mmc->capacity >> 9) - START_TEST_MMC_BLK;
#endif

    printf("Life testing start from block 0x%x len=0x%x block!!!\n",start_t_blk,total_t_len );

    printf("\n\nCurrent Lift information:\n");
    ret = mmc->block_dev.smart_report(curr_device);
    if(ret)
        goto ERR_OUT;

    printf("\n\nNotice -- This testing will damage life time of MMC!!!\n");
    printf("Press any key to continue or esc to exit...\n");

    if (getc() == 0x1B) /* ESC */
        return 0;


    if(old_result){
        printf("\n\nLife testing result cleared.\n");
        memset(result_buf, 0, RESULT_BUF_LEN);
        snprintf(result_buf, RESULT_BUF_LEN, "Life testing result cleared.");
        setenv("em_lf_rst", result_buf);
        saveenv();
    }

    for(k=0; k<test_loop; k++){

        /* Set pattern */
        pattern = pattern_ary[k & 0x3];
        printf("\n\n------- Run %d -------\n",k);
        printf("pattern = 0x%08x\n",pattern);
        for(j=0; j<(MMC_RW_TEST_BLK); j++){
            for(i=0; i<128 ;i++){
                w_buf[(j*128)+i] = pattern;
            }
        }

        act_blk = START_TEST_MMC_BLK;
        remain_t_len = total_t_len;

        while(remain_t_len > MMC_RW_TEST_BLK ){

            act_len = MMC_RW_TEST_BLK;
            if(event_ctl & SHOW_MORE_LOG){
                printf("act_blk=0x%08x len=0x%08x remain=0x%08x\n",
                        act_blk,act_len,remain_t_len);
            }else{
                printf("%s","g");
            }
            //printf("write pattern---\n");
            n = mmc->block_dev.byte_write(curr_device, (u64)act_blk*0x200,act_len*0x200, (void *)w_buf);
            if(n != act_len*0x200){
                printf("%d bytes written: %s\n", n, (n == act_len*0x200) ? "OK" : "ERROR");
                break;
            }

            //printf("clear read buffer---\n");
            for(j=0; j<(MMC_RW_TEST_BLK); j++){
                for(i=0; i<128 ;i++){
                    r_buf[(j*128)+i] = 0;
                }
            }

            //printf("Read data back---\n");
            n = mmc->block_dev.byte_read(curr_device, (u64)act_blk*0x200,act_len*0x200, (void *)r_buf);
            if(n != act_len*0x200){
                printf("%d bytes Read: %s\n", n, (n == act_len*0x200) ? "OK" : "ERROR");
                break;
            }

            //printf("compare data---\n");
            comp_fail = 0;
            for(j=0; j<(MMC_RW_TEST_BLK); j++){
                for(i=0; i<128 ;i++){
#if 0
                    /* error case check */
                    if(j==2 && i==25){
                        r_buf[(j*128)+i] = 0x7878abcd;
                    }

#endif
                    if(r_buf[(j*128)+i] != w_buf[(j*128)+i]){

                            printf("\nerror block[0x%x]---r_buf[%x]=0x%08x w_buf[%x]=0x%08x\n\n",
                                j,(j*128)+i,r_buf[(j*128)+i],(j*128)+i,w_buf[(j*128)+i]);

                            comp_fail = 1;
                            goto FAIL_OUT;
                    }
                }

            }
FAIL_OUT:
            if(comp_fail){
                printf("\nPrint error block 0x%x out.\n",act_blk+j);
                for(i=0; i<(128/8) ;i++){
                    printf("%03x-%03x : %08x %08x %08x %08x %08x %08x %08x %08x\n",
                                (i*32),(i*32)+31,
                                r_buf[(j*128)+(i*8)+0],r_buf[(j*128)+(i*8)+1],
                                r_buf[(j*128)+(i*8)+2],r_buf[(j*128)+(i*8)+3],
                                r_buf[(j*128)+(i*8)+4],r_buf[(j*128)+(i*8)+5],
                                r_buf[(j*128)+(i*8)+6],r_buf[(j*128)+(i*8)+7]);

                }
                memset(result_buf, 0, RESULT_BUF_LEN);
                snprintf(result_buf, RESULT_BUF_LEN, "Life testing fail at block 0x%x ,loop %d/%d!!",
                            act_blk+j,(k+1),test_loop);
                setenv("em_lf_rst", result_buf);
                saveenv();
                goto ERR_OUT;
            }

            act_blk = act_blk+act_len;
            remain_t_len = remain_t_len-act_len;

        }
        memset(result_buf, 0, RESULT_BUF_LEN);
        snprintf(result_buf, RESULT_BUF_LEN, "Life testing %d/%d loop Pass!!",
                    (k+1),test_loop);
        setenv("em_lf_rst", result_buf);
        saveenv();
    }

ERR_OUT:
    return 0;
}


static int do_mmc_read_byte(cmd_tbl_t *cmdtp, int flag,
		       int argc, char * const argv[])
{
	struct mmc *mmc;
	u64 ofs;
	u32 cnt, n;
	void *addr;

	if (argc != 4)
		return CMD_RET_USAGE;

	addr = (void *)simple_strtoul(argv[1], NULL, 16);
	ofs = simple_strtoul(argv[2], NULL, 16);
	cnt = simple_strtoul(argv[3], NULL, 16);

	mmc = init_mmc_device(curr_device, false);
	if (!mmc)
		return CMD_RET_FAILURE;

	printf("\nMMC read: dev # %d, bytes # %d, count %d ... ",
				curr_device, (u32)ofs, cnt);

	n = mmc->block_dev.byte_read(curr_device, ofs,cnt, addr);
	/* flush cache after read */
	flush_cache((ulong)addr, cnt); /* FIXME */
	printf("%d bytes written: %s\n", n, (n == cnt) ? "OK" : "ERROR");

	return (n == cnt) ? CMD_RET_SUCCESS : CMD_RET_FAILURE;
}

static int do_mmc_write_byte(cmd_tbl_t *cmdtp, int flag,
			int argc, char * const argv[])
{
	struct mmc *mmc;
	u64 ofs;
	u32 cnt, n;
	void *addr;

	if (argc != 4)
		return CMD_RET_USAGE;

	addr = (void *)simple_strtoul(argv[1], NULL, 16);
	ofs = simple_strtoul(argv[2], NULL, 16);
	cnt = simple_strtoul(argv[3], NULL, 16);

	mmc = init_mmc_device(curr_device, false);
	if (!mmc)
		return CMD_RET_FAILURE;

	printf("\nMMC write: dev # %d, bytes # %d, count %d ... ",
				curr_device, (u32)ofs, cnt);

#if 0	/* should add to check write protect func */
	if (mmc_getwp(mmc) == 1) {
		printf("Error: card is write protected!\n");
		return CMD_RET_FAILURE;
	}
#endif

	n = mmc->block_dev.byte_write(curr_device, (u32)ofs,cnt, addr);
	printf("%d bytes written: %s\n", n, (n == cnt) ? "OK" : "ERROR");

	return (n == cnt) ? CMD_RET_SUCCESS : CMD_RET_FAILURE;
}

static
ulong mmc_erase_type(struct mmc *mmc, ulong start, lbaint_t blkcnt,unsigned int type)
{
	struct mmc_cmd cmd;
	ulong end;
	int err, start_cmd, end_cmd;

	if (mmc->high_capacity)
		end = start + blkcnt - 1;
	else {
		end = (start + blkcnt - 1) * mmc->write_bl_len;
		start *= mmc->write_bl_len;
	}

	if (IS_SD(mmc)) {
		start_cmd = SD_CMD_ERASE_WR_BLK_START;
		end_cmd = SD_CMD_ERASE_WR_BLK_END;
	} else {
		start_cmd = MMC_CMD_ERASE_GROUP_START;
		end_cmd = MMC_CMD_ERASE_GROUP_END;
	}

	cmd.cmdidx = start_cmd;
	cmd.cmdarg = start;
	cmd.resp_type = MMC_RSP_R1;
	cmd.flags = 0;

	err = mmc_send_cmd(mmc, &cmd, NULL);
	if (err)
		goto err_out;

	cmd.cmdidx = end_cmd;
	cmd.cmdarg = end;

	err = mmc_send_cmd(mmc, &cmd, NULL);
	if (err)
		goto err_out;

	cmd.cmdidx = MMC_CMD_ERASE;
	cmd.cmdarg = type;
	cmd.resp_type = MMC_RSP_R1b;

	err = mmc_send_cmd(mmc, &cmd, NULL);
	if (err)
		goto err_out;

	return 0;

err_out:
	puts("mmc erase failed\n");
	return err;
}


//extern int mmc_send_status(struct mmc *mmc, int timeout,int no_crc);
static
lbaint_t mmc_erase(struct mmc *mmc, unsigned long start, lbaint_t blkcnt,unsigned int type)
{
	int err = 0;
	lbaint_t blk = 0, blk_r = 0;

	if (!mmc)
		return -1;

	if(!(type & TRIM_OPER)){
		if ((start % mmc->erase_grp_size) || (blkcnt % mmc->erase_grp_size))
			printf("\n\nCaution! Your devices Erase group is 0x%x\n"
				"The erase range would be change to 0x%lx~0x%lx\n\n",
				mmc->erase_grp_size, start & ~(mmc->erase_grp_size - 1),
				((start + blkcnt + mmc->erase_grp_size)
				& ~(mmc->erase_grp_size - 1)) - 1);
	}

	while (blk < blkcnt) {

		blk_r = ((blkcnt - blk) > mmc->erase_grp_size) ?
			mmc->erase_grp_size : (blkcnt - blk);
		err = mmc_erase_type(mmc, start + blk, blk_r,type);
		if (err)
			break;

		blk += blk_r;

		/* Waiting for the ready status */
		if (mmc_send_status(mmc, 10000,0))
			return 0;
	}

	return blk;
}


static int do_mmc_report(cmd_tbl_t *cmdtp, int flag,
			 int argc, char * const argv[])
{
	struct mmc *mmc;

	mmc = init_mmc_device(curr_device, false);
	if (!mmc)
		return CMD_RET_FAILURE;

	return mmc->block_dev.smart_report(curr_device);

}


static int do_mmc_wp(cmd_tbl_t *cmdtp, int flag,
			 int argc, char * const argv[])
{
	unsigned long addr = 0;
	unsigned long cnt = 0;
	unsigned char wp_en = 0;

	struct mmc *mmc;

	mmc = init_mmc_device(curr_device, false);

	if (!mmc) {
		printf("no mmc device at slot %x\n", curr_device);
		return CMD_RET_FAILURE;
	}

	if(argc >= 3)
		addr = (unsigned long)simple_strtoul(argv[2], NULL, 16);

	if(argc >= 4)
		cnt = (unsigned long)simple_strtoul(argv[3], NULL, 16);

	if(strcmp(argv[1], "disable_whole") == 0){
		addr = 0x0;
		cnt = mmc->block_dev.lba;
		wp_en = WP_DISABLE;
	}else if(strcmp(argv[1], "enable") == 0){
		wp_en = WP_ENABLE;
	}else if(strcmp(argv[1], "disable") == 0){
		wp_en = WP_DISABLE;
	}else if(strcmp(argv[1], "chks") == 0){
		wp_en = WP_CHECK_STAT;
	}else if(strcmp(argv[1], "chks_all") == 0){
		addr = 0x0;
		cnt = mmc->block_dev.lba;
		wp_en = WP_CHECK_STAT;
	}else if(strcmp(argv[1], "chkt") == 0){
		wp_en = WP_CHECK_TYPE;
	}else if(strcmp(argv[1], "chkt_all") == 0){
		addr = 0x0;
		cnt = mmc->block_dev.lba;
		wp_en = WP_CHECK_TYPE;
	}else if(strcmp(argv[1], "info") == 0){
		cnt = 0x2000;
		wp_en = WP_CHECK_INFO;
	}else{
		return CMD_RET_USAGE;
	}

	return mmc->block_dev.wp_manage(curr_device, addr, cnt, wp_en);

}

extern void eMMC_show_speed(void);
static int do_mmc_speed(cmd_tbl_t *cmdtp, int flag,
			 int argc, char * const argv[])
{
	struct mmc *mmc;

	mmc = init_mmc_device(curr_device, false);

	if (!mmc) {
		printf("no mmc device at slot %x\n", curr_device);
		return CMD_RET_FAILURE;
	}
	eMMC_show_speed();

	return CMD_RET_SUCCESS;

}

extern int export_eMMC_W_R_compare(void);
extern int export_eMMC_W_R_compare_test(unsigned int blk_adr,
			unsigned int blk_cnt,unsigned int loop);
extern int export_eMMC_W_pattern(void);
extern int export_eMMC_chk_pattern(void);

int do_mmcops_compare(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	int ret = -1;
	unsigned int blk_adr = 0;
	unsigned int blk_cnt = 0;
	unsigned int loop = 0;

	if(argc == 1){
		ret = export_eMMC_W_R_compare();
	}else if(strcmp(argv[1], "w_pat") == 0){
		ret = export_eMMC_W_pattern();
	}else if(strcmp(argv[1], "chk_pat") == 0){
		ret = export_eMMC_chk_pattern();
	}else if(argc > 3){
		blk_adr = (unsigned int)simple_strtoul(argv[1], NULL, 16);
		blk_cnt = (unsigned int)simple_strtoul(argv[2], NULL, 16);

		if(argc > 4)
			loop = (unsigned int)simple_strtoul(argv[3], NULL, 16);

		printf("%s(%d)blk_adr=0x%x blk_cnt=0x%x loop=0x%x\n",
			__func__,__LINE__,blk_adr,blk_cnt,loop);

		ret = export_eMMC_W_R_compare_test(blk_adr,blk_cnt,loop);

	}else{
		printf("cmd format no support\n");
		return CMD_RET_FAILURE;
	}

	if(!ret)
		printf("compare pass\n");

	return ret;
}

int do_mmc_debug(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	struct mmc *mmc;

	mmc = init_mmc_device(curr_device, false);

	if (!mmc) {
		printf("no mmc device at slot %x\n", curr_device);
		return CMD_RET_FAILURE;
	}

	extern unsigned int rtkmmc_debug_msg;
	if (argc == 1) {
		if (strcmp(argv[0], "debug") == 0) {
			printf("current debug message output %s\n",
				rtkmmc_debug_msg ? "on" : "off" );
			return 0;
		}
	}
	if (argc == 2) {
		if (strcmp(argv[0], "debug") == 0) {
			printf("debug func from \"%s\" change to ",
				rtkmmc_debug_msg ? "on" : "off" );
			if (strcmp(argv[1], "on") == 0) {
				rtkmmc_debug_msg = 1;
			}
			else if (strcmp(argv[1], "off") == 0) {
				rtkmmc_debug_msg = 0;
			}
			printf("\"%s\"\n", rtkmmc_debug_msg ? "on" : "off" );
			return 0;
		}
	}

	return CMD_RET_USAGE;
}


int do_mmc_extcsd(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	struct mmc *mmc = NULL;
	unsigned int idx = 0;
	unsigned int val = 0;

	mmc = init_mmc_device(curr_device, false);

	if (!mmc) {
		printf("no mmc device at slot %x\n", curr_device);
		return CMD_RET_FAILURE;
	}

	if (argc == 4) {
		if (strcmp(argv[2], "w") == 0) {
			printf("Set ext_CSD\n");
			idx = simple_strtoul(argv[1], NULL, 16);
			val = simple_strtoul(argv[3], NULL, 16);
			printf("idx=0x%02x val=0x%02x\n",idx,val);
			mmc_switch(mmc, EXT_CSD_CMD_SET_NORMAL, idx, val);
			return 0;
		}
	}else{
		printf("Show ext_CSD\n");
		idx = 0xff;
		if (argc == 3) {
			if (strcmp(argv[2], "r") == 0) {
				idx = simple_strtoul(argv[1], NULL, 16);
			}else{
				return CMD_RET_USAGE;
			}
		}
		if(argc == 2){
			idx = simple_strtoul(argv[1], NULL, 16);
		}
		printf("idx=%d(0x%02x)\n",idx,idx);
		export_mmc_show_ext_csd(idx);
		return 0;
	}

	return CMD_RET_USAGE;
}

int do_mmc_chgpart(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	struct mmc *mmc = NULL;
	unsigned int part = 0;

	mmc = init_mmc_device(curr_device, false);

	if (!mmc) {
		printf("no mmc device at slot %x\n", curr_device);
		return CMD_RET_FAILURE;
	}

	if (argc == 2) {
		part = simple_strtoul(argv[1], NULL, 16);
		printf("caange to partition %d\n",part);
		export_mmc_switch_part(part);
		return 0;
	}

	return CMD_RET_USAGE;
}

#endif

#ifdef CONFIG_SUPPORT_EMMC_RPMB
extern int fgets(char *s, int n);
static int confirm_key_prog(unsigned int *key_addr)
{
	char input[10] = {0};
	printf("check your authentication key:");
	printf("%x %x %x %x %x %x %x %x\n", *key_addr, *(key_addr+1), *(key_addr+2),
		*(key_addr+3), *(key_addr+4), *(key_addr+5), *(key_addr+6),
		*(key_addr+7));

	puts("Warning: Programming authentication key can be done only once !\n"
	     "         Use this command only if you are sure of what you are doing,\n"
	     "Really perform the key programming? <yes/N> ");

	fgets(input, 9);
	if (!strncmp(input, "yes", strlen("yes"))) {
	    printf("Authentication key is setting...\n");
		return 1;
	}

	puts("Authentication key programming aborted.\n");
	return 0;
}
static int do_mmcrpmb_key(cmd_tbl_t *cmdtp, int flag,
			  int argc, char * const argv[])
{

	void *key_addr;
	struct mmc *mmc = find_mmc_device(curr_device);

	if (argc != 2)
		return CMD_RET_USAGE;

	key_addr = (void *)simple_strtoul(argv[1], NULL, 16);
	if (!confirm_key_prog((unsigned int *)key_addr))
		return CMD_RET_FAILURE;
	if (mmc_rpmb_set_key(mmc, key_addr)) {
		printf("ERROR - Key already programmed ?\n");
		return CMD_RET_FAILURE;
	}
	return CMD_RET_SUCCESS;
}

static int do_mmcrpmb_read(cmd_tbl_t *cmdtp, int flag,
			   int argc, char * const argv[])
{
	u16 blk, cnt;
	void *addr;
	int n;
	void *key_addr = NULL;
	struct mmc *mmc = find_mmc_device(curr_device);

	if (argc < 4){
		return CMD_RET_USAGE;
	}

	addr = (void *)simple_strtoul(argv[1], NULL, 16);
	blk = simple_strtoul(argv[2], NULL, 16);
	cnt = simple_strtoul(argv[3], NULL, 16);

	if (argc == 5){
		key_addr = (void *)simple_strtoul(argv[4], NULL, 16);
	}

	printf("\nMMC RPMB read: dev # %d, block # %d, count %d ... ",
	       curr_device, blk, cnt);
	n =  mmc_rpmb_read(mmc, addr, blk, cnt, key_addr);

	printf("%d RPMB blocks read: %s\n", n, (n == cnt) ? "OK" : "ERROR");
	if (n != cnt)
		return CMD_RET_FAILURE;
	return CMD_RET_SUCCESS;
}

static int do_mmcrpmb_write(cmd_tbl_t *cmdtp, int flag,
			    int argc, char * const argv[])
{
	u16 blk, cnt;
	void *addr;
	int n;
	void *key_addr;
	struct mmc *mmc = find_mmc_device(curr_device);

	if (argc < 5) {
	    printf("not enough parameter\n");
		return CMD_RET_USAGE;
	}

	addr = (void *)simple_strtoul(argv[1], NULL, 16);
	blk = simple_strtoul(argv[2], NULL, 16);
	cnt = simple_strtoul(argv[3], NULL, 16);
	key_addr = (void *)simple_strtoul(argv[4], NULL, 16);

	printf("\nMMC RPMB write: dev # %d, block # %d, count %d ... ",
	       curr_device, blk, cnt);
	n =  mmc_rpmb_write(mmc, addr, blk, cnt, key_addr);

	printf("%d RPMB blocks written: %s\n", n, (n == cnt) ? "OK" : "ERROR");
	if (n != cnt)
		return CMD_RET_FAILURE;
	return CMD_RET_SUCCESS;
}
static int do_mmcrpmb_counter(cmd_tbl_t *cmdtp, int flag,
			      int argc, char * const argv[])
{
	unsigned long counter;
	struct mmc *mmc = find_mmc_device(curr_device);

	if (mmc_rpmb_get_counter(mmc, &counter))
		return CMD_RET_FAILURE;
	printf("RPMB Write counter= %lx\n", counter);
	return CMD_RET_SUCCESS;
}

static cmd_tbl_t cmd_rpmb[] = {
	U_BOOT_CMD_MKENT(key, 2, 0, do_mmcrpmb_key, "", ""),
	U_BOOT_CMD_MKENT(read, 5, 1, do_mmcrpmb_read, "", ""),
	U_BOOT_CMD_MKENT(write, 5, 1, do_mmcrpmb_write, "", ""),
	U_BOOT_CMD_MKENT(counter, 1, 1, do_mmcrpmb_counter, "", ""),
};

static int do_mmcrpmb(cmd_tbl_t *cmdtp, int flag,
		      int argc, char * const argv[])
{
	cmd_tbl_t *cp;
	struct mmc *mmc;
	int ret;

	cp = find_cmd_tbl(argv[1], cmd_rpmb, ARRAY_SIZE(cmd_rpmb));

	/* Drop the rpmb subcommand */
	argc--;
	argv++;

	if (cp == NULL || argc > cp->maxargs){
		return CMD_RET_USAGE;
	}
	if (flag == CMD_FLAG_REPEAT && !cp->repeatable){
		return CMD_RET_SUCCESS;
	}

	mmc = init_mmc_device(curr_device, false);
	if (!mmc){
		return CMD_RET_FAILURE;
	}


	if (!(mmc->version & MMC_VERSION_MMC)) {
		printf("It is not a EMMC device\n");
		return CMD_RET_FAILURE;
	}
	if (mmc->version < MMC_VERSION_4_4) {
		printf("RPMB not supported before version 4.41\n");
		return CMD_RET_FAILURE;
	}

	/* Switch to the RPMB partition */
	if(mmc_switch_part(curr_device, 3))
		return CMD_RET_FAILURE;

	ret = cp->cmd(cmdtp, flag, argc, argv);

	/* Return to user partition */
	if(mmc_switch_part(curr_device, 0))
		return CMD_RET_FAILURE;

	return ret;

}
#endif	//?#ifdef CONFIG_SUPPORT_EMMC_RPMB


#define EMMC_CMD_VERIFY
#ifdef EMMC_CMD_VERIFY
#define EMMC_TEST_LOG 		1
extern unsigned int rtk_get_rng(void);

static u32  emmc_test_transfer(void *buff, u32 offset, u32 size, u32 write , u32 alert_speed, u32 *time)
{
	struct mmc *mmc;
	u32 tick, elapsed, speed;;
	int rc;

	mmc = init_mmc_device(curr_device, false);
	tick = get_boot_time_us();

	if (write){
		//rc = storage_write(offset, size, (void*)buff);
		//rc = mmc->block_dev.block_write(curr_device, blk,cnt, addr);
		rc = mmc->block_dev.block_write(curr_device,(u32)(offset>>9),(size>>9),(void*)buff);
	}else{
		//rc = storage_read(offset, size, (void*)buff);
		//rc = mmc->block_dev.block_read(curr_device, (u32)blk,cnt, addr);
		rc = mmc->block_dev.block_read(curr_device,(u32)(offset>>9),(size>>9),(void*)buff);
	}
	elapsed = get_boot_time_us() - tick;

	speed = (size/1024)*1000 * 100 /(elapsed/10);
	*(u32*)time = elapsed;

#if (EMMC_TEST_LOG > 0)
	printf("0x%08X(0x%08X) - %8u bytes %s: %s. %8uus elapsed, %8uKbps/sec\n",
			offset, size + offset, size, write > 0 ? "write" : "read ", rc < 0 ? "ER" : "OK", elapsed, speed);
#endif

	if (speed < alert_speed)
	{
		printf("0x%08X(0x%08X) - %7u bytes %s: %s. %3u.%3ums elapsed, %7uKbps/sec\n",
			offset, size + offset, size, write > 0 ? "write" : "read ", rc < 0 ? "ER" : "OK",
			elapsed/1000, elapsed%1000, speed);
	}

	return speed;
}

static
int do_emmc_test(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	static char		*emmc_test_buf;
	//static int		emmc_test_buf_inited = 0;
	char			*endptr;
	u32				i, loop;
	u32				rand_limit;

	u32				e_start;
	u32				e_size;
	u32				e_chunk;
	u32				e_write;

	u32				speed= 0;
	u32				r_speed = 0, r_speed_min = 0xFFFFFFFF, r_speed_max = 0;
	u32				w_speed = 0, w_speed_min = 0xFFFFFFFF, w_speed_max = 0;
	u32				r_count = 0;
	u32				w_count = 0;
	u32				alert_speed;
	u32				chunk_sz;
	struct mmc pemmc;
	u64				offset, size;

	u32				time = 0;
	u32				r_time_min = 0xFFFFFFFF;
	u32				r_time_max = 0;

	u32				w_time_min = 0xFFFFFFFF;
	u32				w_time_max = 0;
	#if (EMMC_TEST_LOG < 1)
	u32				percent;
	#endif

	printf("subcmd_emmc_test start \n");

	//if (argc < 6)
	//	return -1;
	printf("argc:%d\n",argc);

	emmc_test_buf = (char *)0x200000;
	memset(emmc_test_buf, 0xA5, 1024*1024);

    memset((void *)&pemmc, 0, sizeof(pemmc)); // V1 clockworks warning - fix ID 96]

#if 1
	for (i=0; i<argc; i++)
	{
		printf("argv[%u] = %s\n",i,argv[i]);
#if 0
		if (i >= 3)
		{
			//strtoul(argv[i], &endptr, 0);
			 simple_strtoul(argv[i], &endptr, 16);
			if(*endptr != '\0')
				return -1;
		}
#endif
	}
#endif

/**
 *	case1 : sequential read
 *	case2 : sequential write
 *	case3 : random read
 *	case4 : random write
 *	case5 : same address write
 *	case6 : random read & write
 */
	if (argc == 2)
	{
		e_start = 1024 * 1024;
		e_size =  100 * 1024 * 1024;
		e_chunk = 64 * 1024;
	}
	else if (argc == 5)
	{
		e_start =  simple_strtoul(argv[2], &endptr, 16);
		e_size =   simple_strtoul(argv[3], &endptr, 16);
		e_chunk =  simple_strtoul(argv[4], &endptr, 16);
	}
	else
	{
		return -1;
	}

	if((strcmp(argv[1], "case3") == 0)
	|| (strcmp(argv[1], "case4") == 0)
	|| (strcmp(argv[1], "case6") == 0))
	{
		loop = (50*1024*1024)/e_chunk;
		rand_limit = e_size/e_chunk;
	}
	else
	{
		loop = e_size / e_chunk;
		rand_limit = 1024;
	}

	if(strcmp(argv[1], "case1") == 0)	alert_speed = 30*1024;
	if(strcmp(argv[1], "case2") == 0)	alert_speed = 5*1024;
	if(strcmp(argv[1], "case3") == 0)	alert_speed = 1024;
	if(strcmp(argv[1], "case4") == 0)	alert_speed = 20;
	if(strcmp(argv[1], "case5") == 0)	alert_speed = 1024;
	if(strcmp(argv[1], "case6") == 0)	alert_speed = 1024;

	printf("EMMC START ADDR  = 0x%X\n",e_start);
	printf("EMMC TEST  CHUNK = %u Byte\n",e_chunk);
	printf("EMMC TEST  LOOP  = %u\n",loop);
	printf("EMMC TEST  SIZE  = %u Byte\n",e_chunk*loop);
	printf("===============================================================\n");

	if((strcmp(argv[1], "case1") == 0) || (strcmp(argv[1], "case3") == 0))
	{
		e_write = 0;
	}
	else
	{
		e_write = 1;
	}

	//pemmc = get_emmc_info();
	chunk_sz	= pemmc.block_dev.blksz;//pemmc->max_block_size;
	printf("chunk_sz = %d pemmc->block_dev->blksz = %lu \n ",chunk_sz,pemmc.block_dev.blksz);
	r_speed		= 0;
	r_speed_min	= 0xFFFFFFFF;
	r_speed_max	= 0;
	w_speed		= 0;
	w_speed_min	= 0xFFFFFFFF;
	w_speed_max = 0;
	r_count		= 0;
	w_count		= 0;

	for (i=0; i < loop ; i++)
	{
		if((strcmp(argv[1], "case1") == 0) || (strcmp(argv[1], "case2") == 0))
		{
			offset = e_start + (i * e_chunk);
		}
		else if((strcmp(argv[1], "case3") == 0) || (strcmp(argv[1], "case4") == 0) || (strcmp(argv[1], "case6") == 0))
		{
			//offset = e_start + (rand_num(rand_limit) * e_chunk);
			//offset = e_start + ( (random() % rand_limit) * e_chunk);
			offset = e_start + ( (rtk_get_rng() % rand_limit) * e_chunk);
		}
		else if(strcmp(argv[1], "case5") == 0)
		{
			offset = e_start;
		}
		else
		{
			return -1;
		}

		if(strcmp(argv[1], "case6") == 0)
		{
			//e_write = rand_num(1024) & 0x1;
			//e_write = (random() % 1024) & 0x1;
			e_write = (rtk_get_rng() % 1024) & 0x1;
		}
		size = e_chunk;

		if(e_write){
			printf("write to ");
		}else{
			printf("read from ");
		}
		printf("blk:0x%x (offset:0x%x%08x) / len: 0x%x%08x\n",
			(uint32_t)(offset>>9),
			(uint32_t)(offset>>32), (uint32_t)offset,
			(uint32_t)(size>>32), (uint32_t)size);

		speed = emmc_test_transfer(emmc_test_buf,offset,size,e_write, alert_speed, &time);
		#if (EMMC_TEST_LOG < 1)
		percent = (i*100)/((e_size / e_chunk));
		printf("Progress = %u\r",percent);
		#endif
		if (e_write == 0)
		{
			r_speed = r_speed + speed;
			r_count++;

			if (speed < r_speed_min)
				r_speed_min = speed;
			if (speed > r_speed_max)
				r_speed_max = speed;

			if (time >= r_time_max)
				r_time_max = time;
			if (r_time_min >= time)
				r_time_min = time;
		}
		else
		{
			w_speed = w_speed + speed;
			w_count++;

			if (speed < w_speed_min)
				w_speed_min = speed;
			if (speed > w_speed_max)
				w_speed_max = speed;

			if (time >= w_time_max)
				w_time_max = time;
			if (w_time_min >= time)
				w_time_min = time;
		}
	}

	if (r_count)
	{
		printf("EMMC READ  Chunk %7d: %2u.%03u %2u.%03u %2u.%03u Mbyte/sec\n",
			chunk_sz, (r_speed/r_count)/1000, (r_speed/r_count)%1000,
			r_speed_min/1000, r_speed_min%1000, r_speed_max/1000, r_speed_max%1000);
	}

	if (w_count)
	{
		printf("EMMC WRITE Chunk %7d: %2u.%03u %2u.%03u %2u.%03u Mbyte/sec\n",
			chunk_sz, (w_speed/w_count)/1000, (w_speed/w_count)%1000,
			w_speed_min/1000, w_speed_min%1000, w_speed_max/1000, w_speed_max%1000);
	}

	return 0;
}
#endif


static cmd_tbl_t cmd_mmc[] = {
	U_BOOT_CMD_MKENT(info, 3, 0, do_mmcinfo, "", ""),
	U_BOOT_CMD_MKENT(read, 4, 1, do_mmc_read, "", ""),
	U_BOOT_CMD_MKENT(write, 4, 0, do_mmc_write, "", ""),
	U_BOOT_CMD_MKENT(erase, 4, 0, do_mmc_erase, "", ""),
	U_BOOT_CMD_MKENT(rescan, 1, 1, do_mmc_rescan, "", ""),
	U_BOOT_CMD_MKENT(part, 1, 1, do_mmc_part, "", ""),
	U_BOOT_CMD_MKENT(list, 1, 1, do_mmc_list, "", ""),
	U_BOOT_CMD_MKENT(dev, 3, 0, do_mmc_dev, "", ""),

#ifdef CONFIG_SUPPORT_EMMC_RPMB
	U_BOOT_CMD_MKENT(rpmb, CONFIG_SYS_MAXARGS, 1, do_mmcrpmb, "", ""),
#endif

#ifdef CONFIG_RTK_MMC
	U_BOOT_CMD_MKENT(readb, 4, 1, do_mmc_read_byte, "", ""),
	U_BOOT_CMD_MKENT(writeb, 4, 0, do_mmc_write_byte, "", ""),
	U_BOOT_CMD_MKENT(report, 1, 1, do_mmc_report, "", ""),
	U_BOOT_CMD_MKENT(wp, CONFIG_SYS_MAXARGS, 0, do_mmc_wp, "", ""),
	U_BOOT_CMD_MKENT(compare, 5, 1, do_mmcops_compare, "", ""),
	U_BOOT_CMD_MKENT(speed, 1, 1, do_mmc_speed, "", ""),
	U_BOOT_CMD_MKENT(debug, 2, 1, do_mmc_debug, "", ""),
	U_BOOT_CMD_MKENT(extcsd, 4, 1, do_mmc_extcsd, "", ""),
	U_BOOT_CMD_MKENT(chgpart, 2, 1, do_mmc_chgpart, "", ""),
	U_BOOT_CMD_MKENT(lftest, 3, 1, do_mmc_life_test, "", ""),

#endif

#ifdef EMMC_CMD_VERIFY
	U_BOOT_CMD_MKENT(test, 5, 1, do_emmc_test, "", ""),
#endif

#if 0
	U_BOOT_CMD_MKENT(hwpartition, 28, 0, do_mmc_hwpartition, "", ""),

#ifdef CONFIG_SUPPORT_EMMC_BOOT
	U_BOOT_CMD_MKENT(bootbus, 5, 0, do_mmc_bootbus, "", ""),
	U_BOOT_CMD_MKENT(bootpart-resize, 4, 0, do_mmc_boot_resize, "", ""),
	U_BOOT_CMD_MKENT(partconf, 5, 0, do_mmc_partconf, "", ""),
	U_BOOT_CMD_MKENT(rst-function, 3, 0, do_mmc_rst_func, "", ""),
#endif
	U_BOOT_CMD_MKENT(setdsr, 2, 0, do_mmc_setdsr, "", ""),
#endif
};

static int do_mmcops(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	cmd_tbl_t *cp;

	cp = find_cmd_tbl(argv[1], cmd_mmc, ARRAY_SIZE(cmd_mmc));

	/* Drop the mmc command */
	argc--;
	argv++;

	if (cp == NULL || argc > cp->maxargs){
		return CMD_RET_USAGE;
	}
	if (flag == CMD_FLAG_REPEAT && !cp->repeatable){
		return CMD_RET_SUCCESS;
	}

	if (curr_device < 0) {
		if (get_mmc_num() > 0) {
			curr_device = 0;
		} else {
			puts("No MMC device available\n");
			return CMD_RET_FAILURE;
		}
	}
	return cp->cmd(cmdtp, flag, argc, argv);
}

U_BOOT_CMD(
	mmc, CONFIG_SYS_MAXARGS, 1, do_mmcops,
	"MMC sub system",
	"\n"
	"mmc info - display info of the current MMC device\n"
#ifdef CONFIG_RTK_MMC
	"----------\n"
	"mmc debug - show current debug output behavior\n"
	"mmc debug on - enable debug message\n"
	"mmc debug off - disable print debug message\n"
#endif
	"----------\n"
	"mmc read addr blk# cnt\n"
	"mmc write addr blk# cnt\n"
#ifdef CONFIG_RTK_MMC
	"mmc readb addr byte_addr# byte_cnt\n"
	"mmc writeb addr byte_addr# byte_cnt\n"
#endif
	"----------\n"
	"mmc erase blk# cnt\n"
#ifdef CONFIG_RTK_MMC
	"mmc erase secure blk# cnt\n"
	"mmc erase trim blk# cnt\n"
	"----------\n"
	"mmc wp info\n"
	"mmc wp enable start_blk# blk_cnt - enable write-protect setting on mmc device\n"
	"mmc wp disable start_blk# blk_cnt - disable write-protect setting on mmc device\n"
	"mmc wp disable_whole - disable write-protect setting on whole mmc device\n"
	"mmc wp chks start_blk# blk_cnt - To check write-protect status corresponds to the start_addr\n"
	"mmc wp chks_all - To check write-protect status all emmc.\n"
	"mmc wp chkt start_blk# blk_cnt - To check write-protect type corresponds to the start_addr\n"
	"mmc wp chkt_all - To check write-protect type all emmc.\n"
	"----------\n"
	"mmc compare blk# cnt loop - write pattern cnt blocks to blk#, then read back compare\n"
	"mmc compare w_pat - write pattern to mmc\n"
	"mmc compare chk_pat - check pattern of mmc\n"
	"----------\n"
	"mmc extcsd - Display all of ext_csd.\n"
	"mmc extcsd idx# - check item of idx#  from ext_csd.\n"
	"mmc extcsd idx# w val# - set value to assigned idx# to ext_csd.\n"
	"----------\n"
	"mmc chgpart idx# - chang partiton to ixd#; 0:user; 1:boot1; 2:boot2\n"
	"----------\n"
	"mmc lftest result - Display last result of life testing\n"
	"mmc lftest loop# [show] - mmc write/read life testing, #loop is whole mmc testing sycle,\n"
	"                          \"show\" for display cmd information, can skip.\n"


#endif
#ifdef CONFIG_SUPPORT_EMMC_RPMB
	"----------\n"
	"mmc rpmb read addr blk# cnt [address of auth-key] - block size is 256 bytes\n"
	"mmc rpmb write addr blk# cnt <address of auth-key> - block size is 256 bytes\n"
	"mmc rpmb key <address of auth-key> - program the RPMB authentication key.\n"
	"mmc rpmb counter - read the value of the write counter\n"
#endif
	"----------\n"
	"mmc rescan - initialize emmc again \n"
	"mmc part - lists available partition on current mmc device\n"
	"mmc dev [dev] [part] - show or set current mmc device [partition]\n"
	"mmc list - lists available devices\n"
#ifdef CONFIG_RTK_MMC
	"mmc report - report the device life time\n"
	"mmc speed - display current speed that emmc running.\n"
#endif

#ifdef EMMC_CMD_VERIFY
	"mmc test case1-6 start size chunk \n"
	"	case1 : sequential read\n"
	"	case2 : sequential write\n"
	"	case3 : random read\n"
	"	case4 : random write\n"
	"	case5 : same address write\n"
	"	case6 : random read & write\n"
#endif

);


