#include <common.h>
#include <linux/ctype.h>

/***
*long atol(char *nptr) - Convert string to long
*
*Purpose:
*       Converts ASCII string pointed to by nptr to binary.
*       Overflow is not detected.
*
*Entry:
*       nptr = ptr to string to convert
*
*Exit:
*       return long int value of the string
*
*Exceptions:
*       None - overflow is not detected.
*
*******************************************************************************/

long atol(const char *nptr)
{
    int c;              /* current char */
    long total;         /* current total */
    int sign;           /* if '-', then negative, otherwise positive */

    /* skip whitespace */
    while ( isspace((int)(unsigned char)*nptr) )
        ++nptr;

    c = (int)(unsigned char)*nptr++;
    sign = c;           /* save sign indication */
    if (c == '-' || c == '+')
        c = (int)(unsigned char)*nptr++;    /* skip sign */

    total = 0;

    while (isdigit(c)) {
        total = 10 * total + (c - '0');     /* accumulate digit */
        c = (int)(unsigned char)*nptr++;    /* get next char */
    }

    if (sign == '-')
        return -total;
    else
        return total;   /* return result, negated if necessary */
}


/***
*int atoi(char *nptr) - Convert string to long
*
*Purpose:
*       Converts ASCII string pointed to by nptr to binary.
*       Overflow is not detected.  Because of this, we can just use
*       atol().
*
*Entry:
*       nptr = ptr to string to convert
*
*Exit:
*       return int value of the string
*
*Exceptions:
*       None - overflow is not detected.
*
*******************************************************************************/

int atoi(const char *nptr)
{
    return (int)atol(nptr);
}
	
void StrToHex(char *pbDest, char *pbSrc, int nLen)
{
	char h1,h2;
	char s1,s2;
	int i;

	for (i=0; i<nLen; i++)
	{
		h1 = pbSrc[2*i];
		h2 = pbSrc[2*i+1];

		s1 = toupper(h1) - 0x30;
		if (s1 > 9) 
		s1 -= 7;

		s2 = toupper(h2) - 0x30;
		if (s2 > 9) 
		s2 -= 7;

		pbDest[i] = s1*16 + s2;
    }
}


void HexToStr(char *pbDest, char *pbSrc, int nLen)
{
	char ddl,ddh;
	int i;

	for (i=0; i<nLen; i++)
	{
		ddh = 48 + pbSrc[i] / 16;
		ddl = 48 + pbSrc[i] % 16;
		if (ddh > 57) ddh = ddh + 7;
		if (ddl > 57) ddl = ddl + 7;
		pbDest[i*2] = ddh;
		pbDest[i*2+1] = ddl;
	}

	pbDest[nLen*2] = '\0';
}


int check_ip(const char *ip)
{
    if (ip == NULL || strlen(ip) < 7)
    {
        return -1;
    }

    int len = strlen(ip);
    const char *start = ip;
    const char *end = ip + len - 1;
    int ret = 0;
    int dot_count = 0;
    int cur_val = 0;
    int val = -1;

    while (end > start && isspace(*end))
    {
        --end;
    }

    while (start < end && isspace(*start))
    {
        ++start;
    }

    if (start >= end || isdigit(*start) == 0 || isdigit(*end) == 0)
    {
        return -1;
    }

    while (start <= end)
    {
        while (start <= end && '0' <= *start && *start <= '9')
        {
            cur_val = 10 * cur_val + (*start - '0');
            if (cur_val < 0 || cur_val > 255)
            {
                ret = -1;
                break;
            }

            val = cur_val;
            ++start;
        }

        if (start <= end)
        {
            if (*start == '.' && val != -1)
            {
                ++start;
                ++dot_count;
                cur_val = 0;
                val = -1;
            }
            else
            {
                ret = -1;
                break;
            }
        }
        else
        {
            break;
        }
    }

    if (ret == 0 && dot_count != 3)
    {
        ret = -1;
    }

    return ret;
}

int check_mac(const char *mac)
{
    if (mac == NULL || strlen(mac) < 17)
    {
        return -1;
    }

    int len = strlen(mac);
    int ret = 0;
    int num_count = 0;
    int colon_count = 0;
    const char *start = mac;
    const char *end = mac + len - 1;

    while (end > start && isspace(*end))
    {
        --end;
    }

    while (start < end && isspace(*start))
    {
        ++start;
    }

    if ((end - start) != 16 || isxdigit(*start) == 0 || isxdigit(*end) == 0)
    {
        return -1;
    }

    while (start <= end)
    {
        while (start <= end && isxdigit(*start))
        {
            ++start;
            if (++num_count > 2)
            {
                ret = -1;
                break;
            }
        }

        if (start <= end )
        {
            if (*start == ':' && num_count == 2)
            {
                ++colon_count;
                num_count = 0;
                ++start;
            }
            else
            {
                ret = -1;
                break;
            }
        }
        else
        {
            break;
        }
    }

    if (ret == 0 && colon_count != 5)
    {
        ret = -1;
    }

    return ret;
}

