#include <common.h>
#include <command.h>
#include <environment.h>
#include <malloc.h>
#include <linux/string.h>
#include <linux/compiler.h>

#include <mmc.h>

#include "eep_crc32.h"
#include "eeprom.h"

#define FLASH_EEP_START_ADDRESS	0x1400000     // blk_no: 0xA000 ~ 0xBFFF   ( 起始位置 )
#define FLASH_EEP_BLK_SIZE		(0x0200)
#define FLASH_EEP_START_CNT	(FLASH_EEP_START_ADDRESS/FLASH_EEP_BLK_SIZE)
#define FLASH_EEP_BLK_COUNT		1024
#define FLASH_EEP_CRC_SIZE				4
#define FLASH_EEP_VALID_BLOCKSIZE		(FLASH_EEP_BLK_SIZE-FLASH_EEP_CRC_SIZE)

#define EEPDATA_OFFSET_POWERSTATE		(0x0000%FLASH_EEP_BLK_SIZE)
#define EEPDATA_OFFSET_POWERMODE		(0x0289%FLASH_EEP_BLK_SIZE)
#define EEPDATA_OFFSET_PANEL_INDEX		(0x0204%FLASH_EEP_BLK_SIZE)
#define EEPDATA_OFFSET_MAC 				(0x0240%FLASH_EEP_BLK_SIZE)
#define EEPDATA_OFFSET_UARTMODE			(0x0281%FLASH_EEP_BLK_SIZE)
#define EEPDATA_OFFSET_CUSTOMIRCODE		(0x0288%FLASH_EEP_BLK_SIZE)
#define EEPDATA_OFFSET_KEYPAD_DISABLE	(0x0282%FLASH_EEP_BLK_SIZE)
#if defined(Config_Board_TV010_R8)
#define EEPDATA_OFFSET_OLED_SYS_LOCK	(0x02F1%FLASH_EEP_BLK_SIZE)
#endif

static int curr_device = -1;

static INT32 EEP_READALL_FLAG = -1;
static UINT8 EEP_PowerState = 0;
static UINT8 EEP_PowerMode = 0;
static UINT8 EEP_MACADDRESS[6] = {0xff,0xff,0xff,0xff,0xff,0xff};
static UINT32 EEP_PanelIndex = 0;
static UINT8 EEP_UartMode = 1;
static UINT8 EEP_CustomIrCode = 0x0E;
static UINT8 EEP_KEYPAD_DISABLE = 0;
#if defined(Config_Board_TV010_R8)
static UINT8 EEP_OLED_SYS_LOCK = 0;
#endif

static INT32 flashEEPCheckCrc32(UINT8 *blkTmpBuf)
{
	UINT32 crc0,crc1;
	UINT32 i = 0;
	
	crc0 = 0;
	crc1 = 0;
	
	for(i=0;i<4;i++)
	{
		crc0 = (crc0 << 8) & ~0x000000ff; 
		crc0 |= (UINT32)blkTmpBuf[FLASH_EEP_VALID_BLOCKSIZE+i] & 0x000000ff; 
		blkTmpBuf[FLASH_EEP_VALID_BLOCKSIZE+i] = 0xFF;
	}
	
	crc1 = eep_crc32(blkTmpBuf,FLASH_EEP_VALID_BLOCKSIZE);
	
	printf("%s crc0 = 0x%x,crc1 = 0x%x\n",__FUNCTION__,crc0,crc1);
	
	if(crc0 != crc1)
	{
		return -1;
	}
	if(0 == crc0)
	{
		i = 0;
		for(i=0;i<FLASH_EEP_VALID_BLOCKSIZE;i++)
		{
			if(blkTmpBuf[i]!=0)
			{
				break;
			}
		}
		if(i >= FLASH_EEP_VALID_BLOCKSIZE)
		{
			return -1;
		}
	}
	return 1;
}

static INT32 falshEEPReadBlock(UINT32 eep_blk_index, UINT8 *buff)
{
	UINT8 * readBuff = NULL;
	UINT32 n = -1;
	
	if (curr_device < 0) {
		if (get_mmc_num() > 0){
			curr_device = 0;
		}else{
			printf("No MMC device available\n");
			return -1;
		}
	}

	struct mmc *mmc = find_mmc_device(curr_device);
	
	if (!mmc) {
		printf("no mmc device at slot %x\n", curr_device);
		return -1;
	}
	
	readBuff = (UINT8 *)malloc(FLASH_EEP_BLK_SIZE);
	memset(readBuff,0xFF,FLASH_EEP_BLK_SIZE);
	
	mmc_init(mmc);
	
	n = mmc->block_dev.block_read(curr_device, FLASH_EEP_START_CNT+eep_blk_index,
				  1, (void *)readBuff);
	/* flush cache after read */
	flush_cache((ulong)readBuff, 1*FLASH_EEP_BLK_SIZE); /* FIXME */
	
	if((!n)||(-1 == flashEEPCheckCrc32(readBuff)))
	{
		n = -1;
		memset(readBuff,0xFF,FLASH_EEP_BLK_SIZE);
		n = mmc->block_dev.block_read(curr_device, FLASH_EEP_START_CNT+FLASH_EEP_BLK_COUNT+eep_blk_index,
				  1, (void *)readBuff);
		/* flush cache after read */
		flush_cache((ulong)readBuff, 1*FLASH_EEP_BLK_SIZE); /* FIXME */
		if((!n)||(-1 == flashEEPCheckCrc32(readBuff)))
		{
			printf("read mmc block fail\n");
			free(readBuff);
			readBuff = NULL;
			return -1;
		}
	}
	
	memcpy(buff,readBuff,FLASH_EEP_BLK_SIZE);
	free(readBuff);
	readBuff = NULL;
	return 1;
}

static INT32 falshEEPWriteBlock(UINT32 eep_blk_index, UINT8 *buff)
{

	//UINT32 n = -1; //unused
	
	if (curr_device < 0) {
		if (get_mmc_num() > 0){
			curr_device = 0;
		}else{
			printf("No MMC device available\n");
			return -1;
		}
	}

	struct mmc *mmc = find_mmc_device(curr_device);
	
	if (!mmc) {
		printf("no mmc device at slot %x\n", curr_device);
		return -1;
	}
		
	mmc_init(mmc);
	
	UINT32 crc0 = eep_crc32(buff,FLASH_EEP_VALID_BLOCKSIZE);

	buff[FLASH_EEP_VALID_BLOCKSIZE] = (UINT8)((crc0 >> 24)&0x000000ff);
	buff[FLASH_EEP_VALID_BLOCKSIZE+1] = (UINT8)((crc0 >> 16)&0x000000ff);
	buff[FLASH_EEP_VALID_BLOCKSIZE+2] = (UINT8)((crc0 >> 8)&0x000000ff);
	buff[FLASH_EEP_VALID_BLOCKSIZE+3] = (UINT8)(crc0&0x000000ff);
	
	mmc->block_dev.block_write(curr_device, FLASH_EEP_START_CNT+eep_blk_index,1, buff);
	mmc->block_dev.block_write(curr_device, FLASH_EEP_START_CNT+FLASH_EEP_BLK_COUNT+eep_blk_index,1, buff);							  
	return 1;
}

UINT32 ConvertIrCode(UINT8 TmpIrIndex)
{
	UINT32 tmpIRCustomNum = (UINT32)((TmpIrIndex<<8)&0x0000FF00);
	tmpIRCustomNum = (~((tmpIRCustomNum&0x0000FF00)+(UINT32)(TmpIrIndex&0x000000FF)))&0x0000FFFF;
	return tmpIRCustomNum;
}

static void flashEEPReadAll(void)
{
	if (-1 < EEP_READALL_FLAG)
	{
		return;
	}
	
	UINT8 * readBlkBuff = (UINT8 *)malloc(FLASH_EEP_BLK_SIZE);
	memset(readBlkBuff,0xFF,FLASH_EEP_BLK_SIZE);
	if(-1 < falshEEPReadBlock(0,readBlkBuff))
	{
		EEP_PowerState = !!readBlkBuff[EEPDATA_OFFSET_POWERSTATE];
		EEP_READALL_FLAG++;
	}
	
	memset(readBlkBuff,0xFF,FLASH_EEP_BLK_SIZE);
	if(-1 < falshEEPReadBlock(1,readBlkBuff))
	{
		EEP_PowerMode=readBlkBuff[EEPDATA_OFFSET_POWERMODE];
		if(2 < EEP_PowerMode)
		{
			EEP_PowerMode = 0;
		}
		
		EEP_PanelIndex = 0;
		EEP_PanelIndex = readBlkBuff[EEPDATA_OFFSET_PANEL_INDEX];
		EEP_PanelIndex = ((EEP_PanelIndex << 8)&0xffffff00)+readBlkBuff[EEPDATA_OFFSET_PANEL_INDEX+1];
		EEP_PanelIndex = ((EEP_PanelIndex << 8)&0xffffff00)+readBlkBuff[EEPDATA_OFFSET_PANEL_INDEX+2];
		EEP_PanelIndex = ((EEP_PanelIndex << 8)&0xffffff00)+readBlkBuff[EEPDATA_OFFSET_PANEL_INDEX+3];
		
		EEP_UartMode = readBlkBuff[EEPDATA_OFFSET_UARTMODE];
		
		EEP_MACADDRESS[0] = readBlkBuff[EEPDATA_OFFSET_MAC];
		EEP_MACADDRESS[1] = readBlkBuff[EEPDATA_OFFSET_MAC+1];
		EEP_MACADDRESS[2] = readBlkBuff[EEPDATA_OFFSET_MAC+2];
		EEP_MACADDRESS[3] = readBlkBuff[EEPDATA_OFFSET_MAC+3];
		EEP_MACADDRESS[4] = readBlkBuff[EEPDATA_OFFSET_MAC+4];
		EEP_MACADDRESS[5] = readBlkBuff[EEPDATA_OFFSET_MAC+5];
		
		if((0x0E > readBlkBuff[EEPDATA_OFFSET_CUSTOMIRCODE]) && (0x00 < readBlkBuff[EEPDATA_OFFSET_CUSTOMIRCODE]))
		{
			EEP_CustomIrCode = readBlkBuff[EEPDATA_OFFSET_CUSTOMIRCODE];
		}
		
		if(1 == readBlkBuff[EEPDATA_OFFSET_KEYPAD_DISABLE])
		{
			EEP_KEYPAD_DISABLE = 1;
		}

	#if defined(Config_Board_TV010_R8)
		if(1 == readBlkBuff[EEPDATA_OFFSET_OLED_SYS_LOCK])
		{
			EEP_OLED_SYS_LOCK = 1;
		}
	#endif
		EEP_READALL_FLAG++;
	}
	
	free(readBlkBuff);
	readBlkBuff=NULL;
	
	if(0 < EEP_READALL_FLAG)
	{
		EEP_READALL_FLAG = 1;
	}
	
	
}

#if defined(CONFIG_REALTEK_EERPOM_TV010_A)
static UINT8 mac_addr[16];
static UINT8 eep_mac_retOk = 0;

bool Eeprom_SetMacAddressToKernel_Read(void)
{
	//add by skyworth-longchao for set MAC addr
	//unsigned char mac_addr[16];
	//	char set_mac_cmd[100];
	//	char       *raw = NULL;

#if 0 //test use
	mac_addr[0]= 0x00;
	mac_addr[1]= 0x1a;
	mac_addr[2]= 0x9a;
	mac_addr[3]= 0x66;
	mac_addr[4]= 0x88;
	mac_addr[5]= 0xEE;
	setMacAddress(mac_addr);
	printf("\033[1;31m Eeprom_Write hack mac address\033[m\n");
#endif
	memset(mac_addr, 0x00, 6);
	if (getMacAddress(mac_addr) == R_ERR_SUCCESS) {
		printf("a. Eeprom_Read-macaddr is %02x%02x%02x%02x%02x%02x\n",mac_addr[0],mac_addr[1],mac_addr[2],mac_addr[3],mac_addr[4],mac_addr[5]);
		eep_mac_retOk = 1;
	}
	else {
		printf("\033[1;31m a. Eeprom_Read mac fail\033[m\n");
		eep_mac_retOk = 0;
	}

	return 0;	
}

bool Eeprom_SetMacAddressToKernel_Apply(void)
{
	//add by skyworth-longchao for set MAC addr
	//unsigned char mac_addr[16];
	INT8 set_mac_cmd[100];
	//INT8 *raw = NULL; //where used maskd

#if 0 //test use
	mac_addr[0]= 0x00;
	mac_addr[1]= 0x1a;
	mac_addr[2]= 0x9a;
	mac_addr[3]= 0x66;
	mac_addr[4]= 0x88;
	mac_addr[5]= 0xEE;
	setMacAddress(mac_addr);
	printf("\033[1;31m Eeprom_Write hack mac address\033[m\n");
#endif

	//printf("bbb. Eeprom_Read-macaddr is %02x%02x%02x%02x%02x%02x\n",mac_addr[0],mac_addr[1],mac_addr[2],mac_addr[3],mac_addr[4],mac_addr[5]);
	if (eep_mac_retOk) {
		//printf("b. Eeprom_Read-macaddr is %02x%02x%02x%02x%02x%02x\n",mac_addr[0],mac_addr[1],mac_addr[2],mac_addr[3],mac_addr[4],mac_addr[5]);
		if (mac_addr[0] == 0xFF) {
			printf("\033[1;31m mac_addr[0]==0xFF, skip set_mac_cmd\033[m\n");
		}else if ((mac_addr[0] & 0x01) == 0x01){
			printf("\033[1;31m mac_addr[0]:0 == 1, skip set_mac_cmd\033[m\n");
		}else if (mac_addr[0] == 0x00 && mac_addr[1] == 0x00 && mac_addr[2] == 0x00) {
			printf("\033[1;31m mac_addr[0][1][2]==0x00, skip set_mac_cmd\033[m\n");
		}
		else if (mac_addr[0] == mac_addr[1] ) {
			printf("\033[1;31m mac_addr[0] == mac_addr[1], skip set_mac_cmd\033[m\n");
		}
		else {
			sprintf(set_mac_cmd,"%02x:%02x:%02x:%02x:%02x:%02x",mac_addr[0],mac_addr[1],mac_addr[2],mac_addr[3],mac_addr[4],mac_addr[5]);
			printf("\033[1;31m set_mac_cmd is %s \033[m\n",set_mac_cmd);
			//raw = (char*)&set_mac_cmd;
			//if( env_set( "ethaddr", raw, ENV_ATTR_RW, raw, env_mac_s2num ) != OK ) {
			//	printf("\033[1;31m set_mac_cmd fail\033[m\n");
			//}
			if (setenv("ethaddr", set_mac_cmd)  != OK) {
				printf("\033[1;31m set_mac_cmd fail\033[m\n");
			}
		}
	}
	else {
		printf("\033[1;31m b. Eeprom_Read mac fail\033[m\n");
	}

	return 0;	
}
#endif

int setMacAddress(UINT8 *buff)
{
	INT32 i=0;
	UINT8 * readBlkBuff = (UINT8 *)malloc(FLASH_EEP_BLK_SIZE);
	memset(readBlkBuff,0xFF,FLASH_EEP_BLK_SIZE);
	falshEEPReadBlock(1,readBlkBuff);

	for(i=0;i<6;i++)
	{
		EEP_MACADDRESS[i] = buff[i];
		readBlkBuff[EEPDATA_OFFSET_MAC+i] = buff[i];
	}
		
	falshEEPWriteBlock(1,readBlkBuff);
	free(readBlkBuff);
	readBlkBuff = NULL;
	return R_ERR_SUCCESS;
}
int getMacAddress(UINT8 *buff)
{
	flashEEPReadAll();
	memcpy(buff,EEP_MACADDRESS,6);
	return R_ERR_SUCCESS;
}
int setPanelIndex(UINT32 index)
{
	UINT8 * readBlkBuff = (UINT8 *)malloc(FLASH_EEP_BLK_SIZE);
	memset(readBlkBuff,0xFF,FLASH_EEP_BLK_SIZE);
	falshEEPReadBlock(1,readBlkBuff);

	EEP_PanelIndex = index;
	readBlkBuff[EEPDATA_OFFSET_PANEL_INDEX] = (UINT8)((index>>24)&0x000000ff);
	readBlkBuff[EEPDATA_OFFSET_PANEL_INDEX+1] = (UINT8)((index>>16)&0x000000ff);
	readBlkBuff[EEPDATA_OFFSET_PANEL_INDEX+2] = (UINT8)((index>>8)&0x000000ff);
	readBlkBuff[EEPDATA_OFFSET_PANEL_INDEX+3] = (UINT8)((index)&0x000000ff);
		
	falshEEPWriteBlock(1,readBlkBuff);
	free(readBlkBuff);
	readBlkBuff = NULL;
	return R_ERR_SUCCESS;
}
int getPanelIndex(UINT32 *index)
{
	flashEEPReadAll();
	*index = EEP_PanelIndex;
	return R_ERR_SUCCESS;
}
int setUartMode(UINT8 mode)
{
	UINT8 * readBlkBuff = (UINT8 *)malloc(FLASH_EEP_BLK_SIZE);
	memset(readBlkBuff,0xFF,FLASH_EEP_BLK_SIZE);
	falshEEPReadBlock(1,readBlkBuff);

	EEP_UartMode = mode;
	readBlkBuff[EEPDATA_OFFSET_UARTMODE] = mode;
		
	falshEEPWriteBlock(1,readBlkBuff);
	free(readBlkBuff);
	readBlkBuff = NULL;
	return R_ERR_SUCCESS;
}
int getUartMode(UINT8 *mode)
{
	flashEEPReadAll();
	*mode=EEP_UartMode;
	return R_ERR_SUCCESS;
}
int setPowerMode(UINT8 mode)
{
	UINT8 * readBlkBuff = (UINT8 *)malloc(FLASH_EEP_BLK_SIZE);
	memset(readBlkBuff,0xFF,FLASH_EEP_BLK_SIZE);
	falshEEPReadBlock(1,readBlkBuff);

	EEP_PowerMode = mode;
	readBlkBuff[EEPDATA_OFFSET_POWERMODE] = mode;
		
	falshEEPWriteBlock(1,readBlkBuff);
	free(readBlkBuff);
	readBlkBuff = NULL;
	return R_ERR_SUCCESS;
}
int getPowerMode(UINT8 *mode)
{
	flashEEPReadAll();
	*mode = EEP_PowerMode;
	return R_ERR_SUCCESS;
}
int setPowerState(UINT8 mode)
{
	UINT8 * readBlkBuff = (UINT8 *)malloc(FLASH_EEP_BLK_SIZE);
	memset(readBlkBuff,0xFF,FLASH_EEP_BLK_SIZE);
	falshEEPReadBlock(0,readBlkBuff);
	EEP_PowerState = mode;
	readBlkBuff[EEPDATA_OFFSET_POWERSTATE] = mode;
	falshEEPWriteBlock(0,readBlkBuff);
	free(readBlkBuff);
	readBlkBuff = NULL;
	return R_ERR_SUCCESS;
}
int getPowerState(UINT8 *mode)
{
	flashEEPReadAll();
	*mode = EEP_PowerState;
	return R_ERR_SUCCESS;
}
int setCustomIrCode(UINT8 code)
{
	UINT8 * readBlkBuff = (UINT8 *)malloc(FLASH_EEP_BLK_SIZE);
	memset(readBlkBuff,0xFF,FLASH_EEP_BLK_SIZE);
	falshEEPReadBlock(1,readBlkBuff);

	EEP_CustomIrCode = code;
	readBlkBuff[EEPDATA_OFFSET_CUSTOMIRCODE] = code;
		
	falshEEPWriteBlock(1,readBlkBuff);
	free(readBlkBuff);
	readBlkBuff = NULL;
	return R_ERR_SUCCESS;
}
int getCustomIrCode(UINT8 *code)
{
	flashEEPReadAll();
	*code = EEP_CustomIrCode;
	return R_ERR_SUCCESS;
}
int getKeypadDisable(UINT8 *tmpflag)
{
	flashEEPReadAll();
	*tmpflag = EEP_KEYPAD_DISABLE;
	return R_ERR_SUCCESS;
}
#if defined(Config_Board_TV010_R8)
int setOledSysLock(UINT8 flag)
{
	UINT8 * readBlkBuff = (UINT8 *)malloc(FLASH_EEP_BLK_SIZE);
	memset(readBlkBuff,0xFF,FLASH_EEP_BLK_SIZE);
	falshEEPReadBlock(1,readBlkBuff);
	EEP_OLED_SYS_LOCK = flag;
	readBlkBuff[EEPDATA_OFFSET_OLED_SYS_LOCK] = flag;
	falshEEPWriteBlock(1,readBlkBuff);
	free(readBlkBuff);
	readBlkBuff = NULL;
	return R_ERR_SUCCESS;
}
int getOledSysLock(UINT8 *flag)
{
	flashEEPReadAll();
	*flag = EEP_OLED_SYS_LOCK;
	return R_ERR_SUCCESS;
}
#endif
