/*----------------------------------------------------------------------------
*
* The confidential and proprietary information contained in this file may
* only be used by a person authorised under and to the extent permitted
* by a subsisting licensing agreement from ARM Limited or its affiliates.
*
*        (C) COPYRIGHT 2008-2009, 2011-2017 ARM Limited or its affiliates.
*             ALL RIGHTS RESERVED
*             
* This entire notice must be reproduced on all copies of this file
* and copies of this file may only be made by a person if such person is
* permitted to do so under the terms of a subsisting license agreement
* from ARM Limited or its affiliates.
*
* Release   : r0p0-00eac0
*-----------------------------------------------------------------------------*/

/* 

BRIEF DESCRIPTION, PURPOSE AND STRATEGY
=======================================
Tests the correct operation of the global timestamp.

The cycle counter is enabled and then sampled periodically to make sure it is incrementing.



DEBUG HINTS IN CASE OF FAILURE
===========================
a) Check the AXI bus connections as this is the bus that is used to communicate with the Job manager.
b) Check that the GPU's reset signal is not asserted.
c) Check that the GPU input clock is toggling as expected.
d) Check that the DFT signals are disabled.
e) Check that MBIST is disabled.
f) Check the GPU interrupt signal IRQGPU is correctly connected (adv2 should pass if this is okay).
g) Check the GPU's memory accesses. (adv1 should pass if this is okay).

*/


#include "MaliFns.h"
#include "MaliDefines_DVALIN.h"
/*
**************************************************************************
  
  This test starts the global cycle counter and samples it LOOP_COUNT 
  number of times.   The loop value may need to be adjusted to 
  ensure the delay is greater than the timestamp increment period.
  
**************************************************************************
*/

#define CHECK_FOR_INCREMENT         0
#define DO_NOT_CHECK_FOR_INCREMENT  1
#define LOOP_COUNT 0x8


int RunMaliTest_timestamp_DVALIN_part0 (void);
int RunMaliTest_timestamp_DVALIN_part1 (void);
int sample_and_print_global_timestamp(char no_increment);
void clear_irqs_and_set_all_masks (void);
void clear_and_set_masks_for_gpu_irq (void);
void clear_and_set_masks_for_mmu_irq (void);
void clear_and_set_masks_for_job_irq (void);
void mali_delay (void);

static int mali_step;
static unsigned int mali_timestamp_count, mali_timestamp_prev_val;

int RunMaliTest_timestamp_DVALIN (int init) {
  if (init) { mali_step = 0; mali_timestamp_count=0;};
  switch (mali_step) {
    case 0  : return RunMaliTest_timestamp_DVALIN_part0();
    case 1  : return RunMaliTest_timestamp_DVALIN_part1();
    default : return 0;
  };
};

int RunMaliTest_timestamp_DVALIN_part0 (void) {
  int res = 0;
  int gpu_rawstat = Mali_RdReg(GPU_IRQ_RAWSTAT);
  int gpu_status  = Mali_RdReg(GPU_IRQ_STATUS);
  int num_cores, i, core_bitmap;
  int m;
  unsigned int tmp_reg;
  printf("RUNNING TEST: timestamp\n");
  printf("  Purpose: Tests the correct operation of the global timestamp\n");

  // Get current configuration to allow testing all registers
  core_bitmap   = Mali_RdReg(0x0100);

  i = core_bitmap;
  num_cores = 0;
  while(i) {
    num_cores++;
    i &= i-1;
  }

  /* The configuration is now printed out by Mali_Identify()
  printf("Selected configuration:\n");
  printf("  Bus width: %d\n", axi_width);
  printf("  Number of shader cores: %d\n", num_cores);
  printf("  L2 cache size: %d kB\n", ((1<<l2_size)/1024));
  */
  
  Mali_Reset();

  /* 
    ************************************************************
    Sample and printout the global timestamp register 
    ************************************************************
  */
  /* ********************************** */
  
  /* ********************************** */
  clear_irqs_and_set_all_masks();
  /* ********************************** */

  /* ********************************** */
  printf ("Power on the shader cores\n");
  Mali_WrReg(SHADER_PWRON_LO     , 0xFFFFFFFF);
  printf ("Power on the Tiler\n");
  Mali_WrReg(TILER_PWRON_LO      , 0xFFFFFFFF);

  printf ("Poll the power changed all flag until it is set\n");
  (void) Mali_PollReg (GPU_IRQ_RAWSTAT, 10, 1, "GPU_IRQ_RAWSTAT[POWER_CHANGED_ALL]");
  gpu_rawstat = Mali_RdReg(GPU_IRQ_RAWSTAT);
  printf ("GPU IRQ RAWSTAT = %x\n", gpu_rawstat);
  /* ********************************** */

  clear_and_set_masks_for_gpu_irq (); /* clear the power* flags in the gpu irq status registers */

  printf("Now check that the timestamp register is set to 0\n");
  Mali_CheckReg(TIMESTAMP_LO, 0x00000000);
  Mali_CheckReg(TIMESTAMP_HI, 0x00000000);
  
  printf ("Check that the cycle count active bit is low\n");
  tmp_reg = Mali_RdReg(GPU_STATUS); 
  if ((tmp_reg & (1<<6)) != 0)
  {
    res = 1;
    return res;
  }
  
  printf ("Start the cycle counter\n");
  Mali_WrReg (GPU_COMMAND, GPU_COMMAND__CYCLE_COUNT_START);

  /* reset the timestamp previous value tracker */
  mali_timestamp_prev_val = 0;

  printf ("Printout the timestamp making sure it is incrementing\n");
  for (m=0;m<LOOP_COUNT;m++)
  {
    /* wait for a bit */
    mali_delay ();
    res |= sample_and_print_global_timestamp(CHECK_FOR_INCREMENT);
  }
  
  
  printf ("Force a GPU interrupt by writing to the POWER_CHANGED_ALL bit\n");
  Mali_WrReg(GPU_IRQ_RAWSTAT, 0x400);

  mali_step++;
  if (res==0)
  {
    return 255;
  }

  return res;
}

int RunMaliTest_timestamp_DVALIN_part1 (void) {
  int res = 0;
  int gpu_irq_rawstat = Mali_RdReg(GPU_IRQ_RAWSTAT);
  int job_irq_rawstat = Mali_RdReg(JOB_IRQ_RAWSTAT);
  int mmu_irq_rawstat = Mali_RdReg(MMU_IRQ_RAWSTAT);
  int gpu_irq_status = Mali_RdReg(GPU_IRQ_STATUS);
  int job_irq_status = Mali_RdReg(JOB_IRQ_STATUS);
  int mmu_irq_status = Mali_RdReg(MMU_IRQ_STATUS);
   /* DEBUG printout */
   printf ("job_irq_rawstat = %x\n", job_irq_rawstat);
   printf ("mmu_irq_rawstat = %x\n", mmu_irq_rawstat);
   printf ("gpu_irq_status = %x\n", gpu_irq_status);
   printf ("job_irq_status = %x\n", job_irq_status);
   printf ("mmu_irq_status = %x\n", mmu_irq_status);
 
  clear_irqs_and_set_all_masks();
  res |= sample_and_print_global_timestamp(CHECK_FOR_INCREMENT);

  printf ("Stop the cycle counter\n");
  Mali_WrReg (GPU_COMMAND, GPU_COMMAND__CYCLE_COUNT_STOP);

  printf ("Poll the CYCLE_COUNT_ACTIVE flag until it is 0\n");
  (void) Mali_PollReg (GPU_STATUS, 6, 0, "GPU_STATUS[CYCLE_COUNT_ACTIVE]");
  
  printf ("Sample the current value of the timestamp to check the last increment\n");
  res |= sample_and_print_global_timestamp(CHECK_FOR_INCREMENT);
  printf ("Sample the timestamp making sure it has not incremented\n");
  mali_delay();
  res |= sample_and_print_global_timestamp(DO_NOT_CHECK_FOR_INCREMENT);
  mali_delay();
  res |= sample_and_print_global_timestamp(DO_NOT_CHECK_FOR_INCREMENT);
  mali_delay();
  res |= sample_and_print_global_timestamp(DO_NOT_CHECK_FOR_INCREMENT);
  
  
  return res;
}


int sample_and_print_global_timestamp(char no_increment) {
  int res = 0;
  unsigned int my_timestamp_lo    = Mali_RdReg (TIMESTAMP_LO);
  unsigned int my_timestamp_hi    = Mali_RdReg (TIMESTAMP_HI);
  unsigned int my_timestamp_hi_lo;
  if (sizeof(unsigned int) > 32) {
    int shift = 32;
    my_timestamp_hi_lo = my_timestamp_lo | (my_timestamp_hi << shift);
  } else {
    my_timestamp_hi_lo = my_timestamp_lo;
  }

  mali_timestamp_count++;
  if ((my_timestamp_hi_lo > mali_timestamp_prev_val) && no_increment==0)
  {
    printf ("======> %d  GLOBAL TIMESTAMP = %x  hi=%x,lo=%x\n", mali_timestamp_count, my_timestamp_hi_lo, my_timestamp_hi, my_timestamp_lo);
    mali_timestamp_prev_val = my_timestamp_hi_lo;
  }
  else if ((my_timestamp_hi_lo == mali_timestamp_prev_val) && no_increment!=0)
  {
    printf ("======> %d  GLOBAL TIMESTAMP (stopped) = %x  hi=%x,lo=%x\n", mali_timestamp_count, my_timestamp_hi_lo, my_timestamp_hi, my_timestamp_lo);
    mali_timestamp_prev_val = my_timestamp_hi_lo;
  }
  else if (no_increment!=0)
  {
    printf ("======> Error: The GLOBAL TIMESTAMP incremented when it had been stopped %x  hi=%x,lo=%x  prev_value=%x\n", my_timestamp_hi_lo, my_timestamp_hi, my_timestamp_lo, mali_timestamp_prev_val);
    res = 1;
  }
  else
  {
    printf ("======> Error: The GLOBAL TIMESTAMP did not increment %x  hi=%x,lo=%x\n", my_timestamp_hi_lo, my_timestamp_hi, my_timestamp_lo);
    res = 1;
  }
  
  return res;
}

void clear_irqs_and_set_all_masks (void) {
  clear_and_set_masks_for_gpu_irq ();
  clear_and_set_masks_for_mmu_irq ();
  clear_and_set_masks_for_job_irq ();
}

void clear_and_set_masks_for_gpu_irq (void) {
  printf ("Clear the GPU IRQ flags\n");
  Mali_WrReg(GPU_IRQ_CLEAR       , 0xFFFFFFFF);
  printf ("Set the GPU IRQ Mask to all 1s\n");
  Mali_WrReg(GPU_IRQ_MASK        , 0xFFFFFFFF);
}
void clear_and_set_masks_for_mmu_irq (void) {
  printf ("Clear the MMU IRQ flags\n");
  Mali_WrReg(MMU_IRQ_CLEAR       , 0xFFFFFFFF);
  printf ("Set the MMU IRQ Mask to all 1s\n");
  Mali_WrReg(MMU_IRQ_MASK        , 0xFFFFFFFF);
}
void clear_and_set_masks_for_job_irq (void) {
  printf ("Clear the JOB IRQ flags\n");
  Mali_WrReg(JOB_IRQ_CLEAR       , 0xFFFFFFFF);
  printf ("Set the JOB IRQ Mask to all 1s\n");
  Mali_WrReg(JOB_IRQ_MASK        , 0xFFFFFFFF);
}

void mali_delay (void) {
  int n;
  int tmp;
#ifdef MALI_MINIMUM_DELAYS
  for (n=0;n<(2);n++)
#else  
  for (n=0;n<(8192);n++)
#endif  
  {
    /* do something to take up time */
    tmp = tmp + mali_timestamp_prev_val;
  }
}
